/****************************************************************************
	Xiliinx SpartanII ConfigrationController with DLL Driver

	Copyright(c) 2002,2003 Tatsuyuki Satoh

	-The spcisp can configration with EZ-USB JTAG PROBE.
	-Supported SpartanII SerialSlave mode and BounderyScan mode.
	-Xilinx bitstreamfile(*.bit) to Direct configration.
	 (header analyzer is not 100%,some version has problem)
	-if DLL Driver of RAW I/O access was build,
	  any custom Interface Probe can use with this software.

	note:

		use MINW32 to compile.

	Histry
	2002. 3.19 Ver.0.9 T.Satoh
		1st release
	2002. 7.29 Ver.1.00 T.Satoh
		Export probe function to DLL
	2002. 8.20 Ver.1.10 T.Satoh
		Support JTAG DaisyChain & AutoDitect / Select
	2003. 4.16 Ver.1.13
		 Export BounderyScan and CCLK routine
		 Change Display Information
****************************************************************************/

#include <windows.h>
#include <stdio.h>

#include "xilinxbit.h"  /* Xilinx bistream file handler */
#include "bscan.h"      /* JTAG BounderyScan Access     */
#include "devlist.h"    /* BounderyScan Device Library  */
#include "xcclk.h"      /* Xilinx CCLK Configration     */

/* ---------- File Buffer ----------- */
#define TX_MAX 32768
BYTE file_buffer[TX_MAX];

/* -------------------- PROBE hardware dependent interface macros -------------------- */
/*
	External Hardware Driver Function Pointers
	Dynamic Mapped from DLL driver
*/

/* Dynamin functions to DriverDirect */
static DWORD (*jtag_version)(void);            /* Driver version code                                  */
static int (*sh_init)(DWORD,DWORD,char *);	/* initialize PROBE & get name : jtag_init  / ss_init   */
static int (*sh_power)(void);				/* sense TargetPower           : jtag_poewr / ss_power  */
static void (*sh_end)(void);				/* end of PROBE controll       : jtag_end   / ss_end    */

/* Dynamin functions to ConfigrationController */
static int (*sh_start)(DWORD);				/* Start Configration          : bscan_start / xcclk_start */
static int (*sh_write)(BYTE *,DWORD,int);	/* Write ConfigData            : bscan_write / xcclk_write */

/* macros of Hardware Driver Function */
#define PROBE_INIT(PARAM,BAUDRATE,PROBEINFO)  sh_init(PARAM,BAUDRATE,PROBEINFO)
#define PROBE_POWER()      sh_power()
#define PROBE_END()        sh_end()
#define CONFIG_START(ID) sh_start(ID)
#define CONFIG_WRITE(B,S,L) sh_write(B,S,L)

/* -------------------- error status -------------------- */
#define ERROR_FILE   0x8000
#define ERROR_PROBE  0x4000
#define ERROR_TARGET 0x2000

/* -------------------- impact option command -------------------- */
enum{
	OPT_PORT,OPT_BAUD,OPT_MODE,OPT_BITSTREAM,OPT_JEDEC,OPT_DRIVER,OPT_BATCH,OPT_CONTINUE,
	OPT_TARGET
};
static const char *impact_cmd[] =
{
	"-port",
	"-baud",
	"-mode",
	"-bitstream",
	"-jedec",
	"-driver",
	"-f",
	"-continue",
	"-target",
	NULL
};

enum{
	MODE_MS,MODE_BSCAN,MODE_SP,MODE_SS
};

static const char *mode_select[] =
{
	"mserial",		// master serial
	"bscan", 		// bscan
	"sparalell",	// sparallel
	"sserial", 		// sserial
	NULL
};

static const char *md_str[4] = { "MasterSerial","BoundaryScan","SlaveParallel","SlaveSerial"};

/* -------------------- Command Options -------------------- */

static int config_mode;
static int port_addr;
static int baudrate;
static int continus_mode;

//bitstream filename / dll driver name
static char filename[128], dll_name[128] , rsp_name[128];

/* -------------------- BoundaryScan(JTAG) Configration -------------------- */
/*
	Xilinx XAPP188 (v2.0)
*/

/* StartanII JTAG Instruction code */
#define EXTEST   0
#define SAMPLE   1
#define USER1    2
#define USER2    3
#define CFG_OUT  4
#define CFG_IN   5
#define INTEST   7
#define USERCODE 8
#define IDCODE   9
#define HIGHZ    10
#define JSTART   12
#define BYPASS   31

/* StartanII JTAG Configration Command Code */
#define CMD_WCFG   0x01 /* 0001 write config   */
#define CMD_LFRM   0x01 /* 0011 last frame     */
#define CMD_RCFG   0x04 /* 1110 read config    */
#define CMD_START  0x05 /* 1110 startup        */
#define CMD_RCAP   0x06 /* 1110 reset capture  */
#define CMD_RCRC   0x07 /* 1110 reset CRC      */
#define CMD_AHIGH  0x08 /* 1110 GHIGH_B assert */
#define CMD_SWITCH 0x09 /* 1110 SWITCH         */

// 4xBYTE to DWORD conversion (bigendian)
static DWORD byte_to_dword(BYTE *data)
{
	return (DWORD)data[3]*0x1000000 |
		   (DWORD)data[2]*0x10000   |
		   (DWORD)data[1]*0x100     |
		   (DWORD)data[0];
}

// swap bits in DWORD data
static DWORD swap_dword(DWORD value)
{
	DWORD ret = 0;
	int b;
	
	for(b=0;b<32;b++)
	{
		if( value & (1<<b) )
			ret |= 0x80000000>>b;
	}
	return ret;
}

/***************************************************************************/
/* show status register                                                    */
/***************************************************************************/
DWORD jtag_status(void)
{
	BYTE obuf[4];
	DWORD sts;
	static const BYTE sts_data[] = {
		0x28,0x00,0xe0,0x01,
		0x00,0x00,0x00,0x00 };

	static const char *md2_str[2] = { " Hi-z " , "Pullup"};
	static const char *hl_str[2] = { "H","L" };
	static const char *lock_str[2] = { "L","U" };
	static const char *yesno_str[2] = { "YES","NO " };
	static const char *err_str[2] = { "OK ","ERR" };

	// read status
	bscan_reset(6);
	bscan_before_dr = (32-bscan_tdi_bypass)&31;
	bscan_inst(CFG_IN,sts_data,NULL,sizeof(sts_data)*8,1);
	bscan_before_dr = bscan_tdi_bypass;
	bscan_inst(CFG_OUT,NULL,NULL,1,0); // dummy 1 bit ?
	bscan_dr(NULL,obuf,32,1);

	sts = swap_dword(byte_to_dword(obuf));

	printf("FPGA Status :DONE:INIT:    MODE   ( I/O ):GWE:GSR:GHIGH:GTS:IER:LOCK:CRC:\n");
	printf("  %08X  :%s :%s :%s(%s): %s : %s :  %s  : %s : %s :%s%s%s%s:%s:\n",
		sts,
		yesno_str[sts&0x4000?0:1], // DONE
		yesno_str[sts&0x2000?1:0], // INIT
		md_str[(sts>>10)&3],md2_str[sts&0x1000?1:0],
		hl_str[(sts>>5)&1],		// GWE_B
		hl_str[(sts>>8)&1],		// GSR_B
		hl_str[(sts>>9)&1],		// GHIGH_B
		hl_str[(sts>>6)&1],		// GTS_CCFG
		hl_str[(sts>>5)&1],	// IN_ERROR
		lock_str[(sts>>1)&1],lock_str[(sts>>2)&1], // LOCK1 , LOCK2
		lock_str[(sts>>3)&1],lock_str[(sts>>4)&1], // LOCK3 , LOCK4
		err_str[sts&1]);		// CRC ERROR
	return sts;
}

/***************************************************************************/
/* BScan SpartanII shutdown                                                */
/***************************************************************************/
DWORD jtag_shutdown(void)
{
	static const BYTE sht_data0[] =
{
	0x30,0x00,0x80,0x01, // 1.CMD write
	0x00,0x00,0x00,CMD_SWITCH,
	0x30,0x01,0x20,0x01, // 2.COR write
	0x00,0xa0,0xbf,0x2d, //       SHUTDOWN = 1
//	0x00,0x80,0xbf,0x2d, //       SHUTDOWN = 1 (err)

	0x30,0x00,0x80,0x01, // 3.CMD write
	0x00,0x00,0x00,CMD_START,
	0x30,0x00,0x80,0x01, // 4.CMD write
	0x00,0x00,0x00,CMD_RCRC,
	0x00,0x00,0x00,0x00}; // flush PIPE

	static const BYTE sht_data1[] =
{
	0x30,0x00,0x80,0x01, // 4.CMD AHIGH
	0x00,0x00,0x00,CMD_AHIGH,
	0x30,0x01,0x20,0x01, // 2.COR SHUTDOWN = 0
	0x00,0x80,0x3f,0x2d, //       SHUTDOWN = 0
	0x00,0x00,0x00,0x00}; // flush PIPE

	printf("FPGA shutdown\n");

	bscan_reset(6);
	bscan_before_dr = (32-bscan_tdi_bypass)&31;
	bscan_inst(CFG_IN,sht_data0,NULL,sizeof(sht_data0)*8,1);	//   SHUTDOWN
	sleep(50);
	bscan_inst(JSTART,NULL,NULL,16,1);							// 5. 6. TCK 12 cycle
	sleep(50);
	bscan_inst(CFG_IN,sht_data1,NULL,sizeof(sht_data1)*8,1);	// 7. GHIGH command
	sleep(50);
	bscan_before_dr = bscan_tdi_bypass;
}

/***************************************************************************/
/* CONFIG_START() : BoundaryScan start Configration                        */
/***************************************************************************/
int jtag_start(DWORD deviceid)
{
	BYTE ibuf[4];
	DWORD id , sts , rev;
	char *tname;

	// Build JTAG DaisyChain 
	if(!bscan_scan_chain())
		return 1;

	// Target Lockon
	if( bscan_tdo_chips < 0 )
		bscan_tdo_chips = bscan_device_autoditect(deviceid);
	if( (bscan_tdo_chips < 0) || (bscan_tdo_chips >= bscan_num_chain) )
		return 1;
	bscan_select(bscan_tdo_chips);

	// IDCODE read & check
	bscan_reset(6);
	bscan_inst(IDCODE,NULL,ibuf,32,1);
	id = byte_to_dword(ibuf);
	switch(id & 0x0fffffff)
	{
	case 0x00608093: tname = "XC2S15"; break;
	case 0x0060c093: tname = "XC2S30"; break;
	case 0x00610093: tname = "XC2S50"; break;
	case 0x00614093: tname = "XC2S100"; break;
	case 0x00618093: tname = "XC2S150"; break;
	case 0x0061c093: tname = "XC2S200"; break;
	default: tname = "unknown";
	}
	printf("Device ID code %08XH : %s Rev.%d\n",id,tname,id>>28);

	if( (id == 0xffffffff || id == 0x00000000))
	{
		printf("JTAG DEVICE error\n");
		return 1;
	}

	/* do shutdown */
#if 0
	jtag_status();
#endif
	jtag_shutdown();
	jtag_status();

	//Enter CFG_IN phase
	bscan_reset(6);
	bscan_before_dr = (32-bscan_tdi_bypass)&31;
	bscan_inst(CFG_IN,NULL,NULL,0,0);
	bscan_before_dr = bscan_tdi_bypass;

	return 0;
}

/***************************************************************************/
/* CONFIG_WRITE() : BoundaryScan write Configration data                   */
/***************************************************************************/
int jtag_write(BYTE *data,DWORD size,int last)
{
	//phase 9,10,11
	bscan_dr(data,NULL,size*8,last);

	//ŌH
	if(last)
	{
		//phase 12,13,14,15,16,17,18
		//phase 19 RUN idle
		bscan_inst(JSTART,NULL,NULL,14,1);
		// show status
		jtag_status();
	}
	return 0;
}

/***************************************************************************/
/* Dynamic DeviceDriver for Windows DLL                                    */
/***************************************************************************/

static HINSTANCE dllHandle;

// get DLL function
static int get_dll_func(void *func_ptr,char *name)
{
	*(void **)func_ptr = (void *)GetProcAddress(dllHandle,name);

	if( *(void **)func_ptr == NULL)
	{
		printf("DLL function '%s' is not found\n",name);
		return 0;
	}
	return 1;
}

static int dll_setup(char *dll_name,int mode)
{
	char *name;
	DWORD ver;

	/* Load DLL */
	dllHandle = LoadLibraryEx(dll_name,NULL,0);
	if(dllHandle == NULL)
	{
		printf("Can't open '%s.DLL'\n",dll_name);
		return 0;
	}

	/* */
	switch(mode)
	{
	case MODE_BSCAN:	// JTAG BoundaryScan
		sh_start = jtag_start;
		sh_write = jtag_write;
		if( 
			get_dll_func(&sh_init,"jtag_init") &
			get_dll_func(&sh_power,"jtag_power") &
			get_dll_func(&jtag_txrx,"jtag_txrx") &
			get_dll_func(&sh_end,"jtag_end")
		)
		{
			if(!get_dll_func(&jtag_version,"jtag_version"))
				jtag_version = NULL;
			return 1; /* okay */
		}
	case MODE_SS:	// SlaveSerial interface
		sh_start = xcclk_start;
		sh_write = xcclk_write;
		if(
			get_dll_func(&sh_init,"ss_init") &
			get_dll_func(&sh_power,"ss_power") &
			get_dll_func(&xcclk_tx,"ss_tx") &
			get_dll_func(&xcclk_program,"ss_program") &
			get_dll_func(&xcclk_done,"ss_done") &
			get_dll_func(&sh_end,"ss_end")
		)
			if(!get_dll_func(&jtag_version,"jtag_version"))
				jtag_version = NULL;
			return 1; /* okay */
	case MODE_SP:	// SlaveParallel interface
		break;
	}

	/* NG */
	FreeLibrary(dllHandle);
	dllHandle = NULL;
	return 0;
}

static void dll_end(void)
{
	if( dllHandle != NULL)
		FreeLibrary(dllHandle);
}

/***************************************************************************/
/* optrion command handling                                                */
/***************************************************************************/
int get_str_cmd(const char **cmd_list , char *str)
{
	int i;

	for(i=0 ; cmd_list[i] != NULL ; i++)
	{
		if(!strcmp(str,cmd_list[i]))
			return i;
	}
	// not found
	return -1;
}

/*
	analyze one of option
	return == No. of parameter used
*/
int analyze_option(char *p1,char *p2)
{
	int param;
	char dummy[1] = {0x00};

	if(p1==NULL)
		return 0;

	strlwr(p1);
	if(p2==NULL)
		p2 = dummy;
	else
		strlwr(p2);

/* printf("OPTION '%s' '%s'\n",p1,p2); */

	if(*p1 == '-')
	{
		/* search option code */
		switch(get_str_cmd(impact_cmd , p1))
		{
		case OPT_PORT: // -port auto
			sscanf(p2,"%X",&port_addr);
			return 2;
		case OPT_BAUD: // -baud auto
			sscanf(p2,"%d",&baudrate);
			return 2;
		case OPT_MODE: // -mode [bscan/sserial/sparalell]
			param = get_str_cmd(mode_select , p2);
			if( param >=0)
				config_mode = param;
			return 2;
		case OPT_BITSTREAM: // -bitstream
			return 1;
		case OPT_JEDEC:	// -jedec
			return 1;
		case OPT_DRIVER: // -driver dllname
			strcpy(dll_name,p2);
			return 2;
		case OPT_BATCH:	// -f rsp_filename
			strcpy(rsp_name,p2);
			return 2;
		case OPT_CONTINUE:	// -continue
			continus_mode = 1;
			return 1;
		case OPT_TARGET:	// -target [number of TDI chain devices]
			sscanf(p2,"%d",&bscan_tdo_chips);
			return 2;
		default:
		}
	}
	else
	{
		// bitstream/jedec_filename
		strcpy(filename,p1);
	}
	return 1;
}

void file_option(char *filepath)
{
	FILE *hndl;
	char buf[256] , *p1,*p2;

	// file open
	hndl = fopen(filepath,"rt");
	if( hndl==NULL)
		return;

	while( fgets(buf,255,hndl) != NULL)
	{
		p1 = strtok(buf," \n");
		p2 = strtok(NULL," \n");
		analyze_option(p1,p2);
	}
	fclose(hndl);
}

/***************************************************************************/
/* MAIN                                                                    */
/***************************************************************************/
int main(int argc,char *argv[])
{
	BIT_FILE *hndl = NULL;
	DWORD file_size;
	char probe_name[128];
	DWORD file_pos,tx_size;
	int last_block;
	int c , param;
	char *p;
	int err = 0;
	int ver_mjr , ver_mnr;
	int option_f;
	char last_date[64],last_time[64];

	// ^Cg\
	printf("\n\
Xilinx SpartanII Configrater with Cypress EZ-USB Ver.1.13\n\
Copyrght(c) 2002,2003 Tatsuuki Satoh\n");

	// 
	if( argc<2)
	{
		printf("\n\
useage: spcusb [Xilinx_bit_file] [options]\n\
  options :\n\
       -port [address]  : port address\n\
       -baud [bps]      : baud rate\n\
       -mode bscan      : BoundaryScan  (JTAG) mode\n\
       -mode sserlal    : SlaveSerial   (CCLK) mode\n\
       -mode sparalell  : SlaveParalell (CCLK) mode\n\
       -bitstream       : [future]\n\
       -jedec           : [future]\n\
       -driver [dllname]: DLL driver slelect\n\
       -continue        : continus mode\n\
       -target [nums]   : number of chained device in TDO side\n\
       -f [rspname]     : use option in file\n\
");
		exit(1);
	}

	// R}hIvV̏
	port_addr = 0;
	baudrate = 0;
	config_mode = MODE_BSCAN;
	continus_mode = 0;
	bscan_tdo_chips = -1;		// Target AutoDitect

	// NIvV
	strcpy(dll_name,"spc_isp"); // default DLL name
	filename[0] = 0x00;
	rsp_name[0] = 0x00;
	option_f = 0;
	for( c=1 ; c<argc ; )
		c = c + analyze_option(argv[c],(c+1)<argc ? argv[c+1] : NULL);

	// .rspt@CLȂAt@CIvVǉ
	if(strstr(rsp_name,".rsp"))
	{
		/* wt@Cǂݍ */
		file_option(rsp_name);
	}

	/* ftH[grspt@CŌɏ㏑ */
	file_option("spcisp.rsp");

	//t@C.bitgqǉ
	if(!strstr(filename,".bit"))
		strcat(filename,".bit");

	err = 0;
	// hCo[DLL̃[h
	if( !dll_setup(dll_name,config_mode) )
		err = 6;

	// Probe Hardware Initialize
	if(!err)
	{
		err = PROBE_INIT(port_addr,baudrate,probe_name);
		switch(err)
		{
		case 0:
			printf("-------------------- Probe / Driver Information ---------------------\n");
			printf("Name    '%s'\n",probe_name);
			if(jtag_version)
			{
				DWORD version = jtag_version();
				printf("Version %2X.%04X\n",version>>16,version&0xffff);
			}
			else
				printf(" Version : Unknown\n");
			break;
		case 2: printf("NO Probe\n"); break;
		case 4: printf("NO Target\n"); break;
		}
	}

	printf("++++++++++++++++++++++ File / Mode Information ++++++++++++++++++++++\n");

	last_date[0] = last_time[0] = 0x00;
again:
	if(err<2)
		err = 0;

	if(!err)
	{
		if( continus_mode )
		{
			printf("Waiting for update '%s'\n",filename);
			while(1)
			{
				hndl = bitfile_open(filename);
				if(hndl == NULL)
					continue;
				if( !strcmp(last_date,hndl->date) && !strcmp(last_time,hndl->time))
				{
					bitfile_close(hndl);
					hndl = NULL;
					sleep(100);
					continue;
				}
				break;
			}
		}
		else
		{
			// t@C̃I[v
			hndl = bitfile_open(filename);
			if( hndl == NULL )
			{
				printf("Can't open bitstream file '%s'\n",filename);
				err = 1;
			}
		}
	}
	// bitstream t@C̏\
	if(!err)
	{
		file_size = hndl->bit_size;
		printf("MODE          %s\n",md_str[config_mode]);
		printf("ConfigFile    '%s'\n",filename);
		printf(" -Netname     '%s'\n"   ,hndl->netname);
		printf(" -TargetDeive '%s'[%08X]\n"   ,hndl->targetname,hndl->deviceid);
		printf(" -TimeStamp   '%s %s'\n",hndl->date,hndl->time);
		printf(" -ConfigSize  %dbits(%dbytes)\n",file_size*8,file_size);
		printf("---------------------------------------------------------------\n");
	}

	// Target Power Check
	if(!err)
	{
		if( !PROBE_POWER() )
		{
			printf("Target power down\n");
			err = 1;
		}
	}

	// Configration Jn
	if(!err)
	{
		err |= CONFIG_START(hndl->deviceid);
		printf("Start Configraion\r");
	}

	// RtBO[Vf[^̑M
	if(!err)
	{
		for(file_pos=0; err==0 && file_pos<file_size ; file_pos += tx_size)
		{
			// ŏIubNH
			last_block = (file_size <= file_pos+TX_MAX);
			// ]
			tx_size = last_block ? (file_size - file_pos) : TX_MAX;

			//\
			printf("Configration bit %7d to -%7d\r",file_pos*8,(file_pos+tx_size)*8-1);

			// bitstreamt@C烊[h
			if( !bitfile_read(hndl,file_buffer,tx_size) )
			{
				printf("File Read Error");
				err |= 1;
				break;
			}
			// PubN̑Mf[^M{݊`FbNiŏIubNj
			err |= CONFIG_WRITE(file_buffer,tx_size,last_block);
		}
	}

	//RtBOt@C̃N[Y
	if(hndl)
	{
		if(!err)
		{
			strcpy(last_date,hndl->date);
			strcpy(last_time,hndl->time);
		}
		bitfile_close(hndl);
		hndl = NULL;
	}

	// ʕ\
	printf("%s\n" , err? "error!" : "Complete!");

	if( continus_mode && err<2)
	{
		if(err)
			sleep(1000);
		goto again;
	}

	// finish PROBE controll
	if(err < 6)
		PROBE_END();

	// DLL close
	dll_end();

	exit(err);
}

